#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Initialize DART
# Adrien Crovato

def initDart(params, scenario='aerodynamic', task='analysis'):
    """Initlialize DART for various API

    Parameters
    ----------
    params : dict
        Dictionary of parameters to configure DART
    scenario : string, optional
        Type of scenario (available: aerodynamic, aerostructural) (default: aerodynamic)
    task : string, optional
        Type of task (available: analysis, optimization) (default: analysis)

    Returns
    -------
    _dim : int
        Dimension of the problem
    _qinf : float
        Freestream dynamic pressure
    _msh : tbox.MshData object
        Mesh data structure
    _wrtr : tbox.MshExport object
        Mesh writer
    _mrf : tbox.MshDeform object
        Mesh morpher
    _pbl : dart.Probem object
        Problem definition
    _bnd : Dart.Body object
        Body of interest
    _sol : dart.Newton object
        Direct (Newton) solver
    _adj : dart.Adjoint object
        Adjoint solver
    """
    # Imports
    import math
    import fwk.wutils as wu
    import tbox
    import tbox.gmsh as gmsh
    import dart

    # Basic checks and config
    # dimension
    if params['Dim'] != 2 and params['Dim'] != 3:
        raise RuntimeError('Problem dimension should be 2 or 3, but ' + params['Dim'] + ' was given!\n')
    else:
        _dim = params['Dim']
    # aoa and aos
    if 'AoA' in params:
        alpha = params['AoA'] * math.pi / 180
    else:
        alpha = 0
    if _dim == 2:
        if 'AoS' in params and params['AoS'] != 0:
            raise RuntimeError('Angle of sideslip (AoS) should be zero for 2D problems!\n')
        else:
            beta = 0
        if 'Symmetry' in params:
            raise RuntimeError('Symmetry boundary condition cannot be used for 2D problems!\n')
    else:
        if 'AoS' in params:
            if params['AoS'] != 0 and 'Symmetry' in params:
                raise RuntimeError('Symmetry boundary condition cannot be used with nonzero angle of sideslip (AoS)!\n')
            beta = params['AoS'] * math.pi / 180
        else:
            beta = 0
    # save format
    if params['Format'] == 'vtk':
        try:
            import tboxVtk
            Writer = tboxVtk.VtkExport
            print('VTK libraries found! Results will be saved in VTK format.\n')
        except:
            Writer = tbox.GmshExport
            print('VTK libraries not found! Results will be saved in gmsh format.\n')
    else:
        Writer = tbox.GmshExport
        print('Results will be saved in gmsh format.\n')
    # number of threads
    if 'Threads' in params:
        nthrd = params['Threads']
    else:
        nthrd = 1
    wu.initMKL(nthrd) # initialize threading layer and number of threads
    # verbosity
    if 'Verb' in params:
        verb = params['Verb']
    else:
        verb = 0
    # scenario and task type
    if scenario != 'aerodynamic' and scenario != 'aerostructural':
        raise RuntimeError('Scenario should be aerodynamic or aerostructural, but "' + scenario + '" was given!\n')
    if task != 'analysis' and task != 'optimization':
        raise RuntimeError('Task should be analysis or optimization, but "' + task + '" was given!\n')
    # dynamic pressure
    if scenario == 'aerostructural':
        _qinf = params['Q_inf']
    else:
        _qinf = 0

    # Mesh creation
    _msh = gmsh.MeshLoader(params['File']).execute(**params['Pars'])
    # add the wake
    if _dim == 2:
        mshCrck = tbox.MshCrack(_msh, _dim)
        mshCrck.setCrack(params['Wake'])
        mshCrck.addBoundaries([params['Fluid'], params['Farfield'][-1], params['Wing']])
        mshCrck.run()
    else:
        for i in range(len(params['Wakes'])):
            mshCrck = tbox.MshCrack(_msh, _dim)
            mshCrck.setCrack(params['Wakes'][i])
            mshCrck.addBoundaries([params['Fluid'], params['Farfield'][-1], params['Wings'][i]])
            if 'Fuselage' in params:
                mshCrck.addBoundaries([params['Fuselage']])
            if 'Symmetry' in params:
                mshCrck.addBoundaries([params['Symmetry']])
            if 'WakeTips' in params:
                mshCrck.setExcluded(params['WakeTips'][i]) # 2.5D computations
            mshCrck.run()
    # save the updated mesh in native (gmsh) format and then set the writer to the requested format
    tbox.GmshExport(_msh).save(_msh.name)
    del mshCrck
    _wrtr = Writer(_msh)
    print('\n')

    # Mesh morpher creation
    if scenario == 'aerostructural' or task == 'optimization':
        _mrf = tbox.MshDeform(_msh, _dim, nthrds=nthrd)
        _mrf.setField(params['Fluid'])
        _mrf.addFixed(params['Farfield'])
        if _dim == 2:
            _mrf.addMoving([params['Wing']])
            _mrf.addInternal([params['Wake'], params['Wake']+'_'])
        else:
            if 'Fuselage' in params:
                _mrf.addFixed(params['Fuselage'])
            _mrf.setSymmetry(params['Symmetry'], 1)
            for i in range(len(params['Wings'])):
                if i == 0:
                    _mrf.addMoving([params['Wings'][i]]) # TODO body of interest (FSI/OPT) is always first given body
                else:
                    _mrf.addFixed([params['Wings'][i]])
                _mrf.addInternal([params['Wakes'][i], params['Wakes'][i]+'_'])
        _mrf.initialize()
    else:
        _mrf = None

    # Problem creation
    _pbl = dart.Problem(_msh, _dim, alpha, beta, params['M_inf'], params['S_ref'], params['c_ref'], params['x_ref'], params['y_ref'], params['z_ref'])
    # add the field
    _pbl.set(dart.Fluid(_msh, params['Fluid'], params['M_inf'], _dim, alpha, beta))
    # add the initial condition
    _pbl.set(dart.Initial(_msh, params['Fluid'], _dim, alpha, beta))
    # add farfield boundary conditions (Neumann only, a DOF will be pinned automatically)
    for i in range (len(params['Farfield'])):
        _pbl.add(dart.Freestream(_msh, params['Farfield'][i], _dim, alpha, beta))
    # add solid boundaries
    if _dim == 2:
        bnd = dart.Body(_msh, [params['Wing'], params['Fluid']])
        _bnd = bnd
        _pbl.add(bnd)
    else:
        _bnd = None
        for bd in params['Wings']:
            bnd = dart.Body(_msh, [bd, params['Fluid']])
            if _bnd is None:
                _bnd = bnd # TODO body of interest (FSI/OPT) is always first given body
            _pbl.add(bnd)
        if 'Fuselage' in params:
            bnd = dart.Body(_msh, [params['Fuselage'], params['Fluid']])
            _pbl.add(bnd)
    # add Wake/Kutta boundary conditions
    if _dim == 2:
        _pbl.add(dart.Wake(_msh, [params['Wake'], params['Wake']+'_', params['Fluid']]))
        _pbl.add(dart.Kutta(_msh, [params['Te'], params['Wake']+'_', params['Wing'], params['Fluid']]))
    else:
        for i in range(len(params['Wakes'])):
            _pbl.add(dart.Wake(_msh, [params['Wakes'][i], params['Wakes'][i]+'_', params['Fluid'], params['TeTips'][i]]))

    # Direct (forward) solver creation
    # NB: more solvers/options are available but we restrict the user's choice to the most efficient ones
    # initialize the linear (inner) solver
    if params['LSolver'] == 'PARDISO':
        linsol = tbox.Pardiso()
    elif params['LSolver'] == 'MUMPS':
        linsol = tbox.Mumps()
    elif params['LSolver'] == 'GMRES':
        linsol = tbox.Gmres()
        linsol.setFillFactor(params['G_fill']) if 'G_fill' in params else linsol.setFillFactor(2)
        linsol.setRestart(params['G_restart']) if 'G_restart' in params else linsol.setRestart(50)
        linsol.setTolerance(params['G_tol']) if 'G_tol' in params else linsol.setTolerance(1e-5)
    else:
        raise RuntimeError('Available linear solvers: PARDISO, MUMPS or GMRES, but ' + params['LSolver'] + ' was given!\n')
    # initialize the nonlinear (outer) solver
    _sol = dart.Newton(_pbl, linsol, rTol=params['Rel_tol'], aTol=params['Abs_tol'], mIt=params['Max_it'], nthrds=nthrd, vrb=verb)

    # Adjoint (reverse) solver creation
    if task == 'optimization':
        if params['LSolver'] == 'PARDISO':
            alinsol = tbox.Pardiso()
        elif params['LSolver'] == 'MUMPS':
            alinsol = tbox.Mumps()
        else:
            alinsol = tbox.Gmres()
            alinsol.setFillFactor(params['G_fill']) if 'G_fill' in params else alinsol.setFillFactor(2)
            alinsol.setRestart(params['G_restart']) if 'G_restart' in params else alinsol.setRestart(50)
            alinsol.setTolerance(1e-12)
        _adj = dart.Adjoint(_sol, _mrf, alinsol)
    else:
        _adj = None

    # Return
    return _dim, _qinf, _msh, _wrtr, _mrf, _pbl, _bnd, _sol, _adj
